
#include "dis_avr.h"

static const char *reg_names[] = 
{
	"r0",   "r1",   "r2",  "r3",  "r4",  "r5",  "r6",  "r7",
	"r8",   "r9",  "r10", "r11", "r12", "r13", "r14", "r15",
	"r16",  "r17",  "r18", "r19", "r20", "r21", "r22", "r23",
	"r24",  "r25",  "r26", "r27", "r28", "r29", "r30", "r31"
};

#define OPCODE_NUMBER 118

avr_opcode opcode_table[OPCODE_NUMBER] =
//avr_opcode opcode_table[]=
{
	//{,,,,,,,,,,,}, 
	{0xFC00,0x1C00,"ADC",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xFC00,0x0C00,"ADD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F, 0},
	{0xFF00,0x9600,"ADIW",		0,	2,	2,	0,	P_Rd,	0x0030,	P_K,	0x00CF,	0},
	{0xFC00,0x2000,"AND",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xF000,0x7000,"ANDI",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_K,	0x0F0F,	0},
	{0xFE0F,0x9405,"ASR",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFF8F,0x9488,"BCLR",		0,	2,	1,	0,	P_s,	0x0070,	P_NONE,	0,		0}, 
	{0xFE08,0xF800,"BLD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_b,	0x0007,	0},
	{0xFC00,0xF400,"BRBC",		0,	2,	2,	0,	P_s,	0x0007,	P_k,	0x03F8,	0},
	{0xFC00,0xF000,"BRBS",		0,	2,	2,	0,	P_s,	0x0007,	P_k,	0x03F8,	0},
	{0xFFFF,0x9598,"BREAK",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFF8F,0x9408,"BSET",		0,	2,	1,	0,	P_s,	0x0070,	P_NONE, 0,		0},
	{0xFE00,0xFA00,"BST",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_b,	0x0007,	0},
	{0xFE0E,0x940E,"CALL",		0,	4,	1,	0,	P_kk,	0x01F1,	P_NONE, 0,		0},
	{0xFF00,0x9800,"CBI",		0,	2,	2,	0,	P_A,	0x00F8,	P_b,	0x0007,	0},
	{0xFFFF,0x9488,"CLC",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x94D8,"CLH",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x94F8,"CLI",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x94A8,"CLN",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFC00,0x2400,"CLR",		0,	2,	1,	0,	P_Rd,	0x03FF,	P_NONE,	0,		0},
	{0xFFFF,0x94C8,"CLS",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x94E8,"CLT",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x94D8,"CLV",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9498,"CLZ",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFE0F,0x9400,"COM",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFC00,0x1400,"CP",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xFC00,0x0400,"CPC",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xF000,0x3000,"CPI",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_K,	0x0F0F,	0},
	{0xFC00,0x1000,"CPSE",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xFE0F,0x940A,"DEC",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFF0F,0x940B,"DES",		0,	2,	1,	0,	P_K,	0x00F0,	P_NONE,	0,		0},
	{0xFFFF,0x9519,"EICALL",	0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9419,"EIJMP",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x95D8,"ELPM",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFE0F,0x9006,"ELPM",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Z,	0,		0},//tipa 2 parama no realjno 2 param dlja ponta tolko
	{0xFE0F,0x9007,"ELPM",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Zp,	0,		0},
	{0xFC00,0x2400,"EOR",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xFF88,0x0308,"FMUL",		0,	2,	2,	0,	P_Rd,	0x0070, P_Rr,	0x0007,	0},
	{0xFF88,0x0380,"FMULS",		0,	2,	2,	0,	P_Rd,	0x0070,	P_Rr,	0x0007,	0},
	{0xFF88,0x0388,"FMULSU",	0,	2,	2,	0,	P_Rd,	0x0070,	P_Rr,	0x0007, 0},
	{0xFFFF,0x9509,"ICALL",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9409,"IJMP",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xF800,0xB000,"IN",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFE0F,0x9403,"INC",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFE0E,0x940C,"JMP",		0,	4,	1,	0,	P_kk,	0x01F1,	P_NONE, 0,		0},
	{0xFE0F,0x900C,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_X,	0,		0},
	{0xFE0F,0x900D,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Xp,	0,		0},
	{0xFE0F,0x900E,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_nX,	0,		0},
	{0xFE0F,0x8008,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Y,	0,		0},
	{0xFE0F,0x9009,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Yp,	0,		0},
	{0xFE0F,0x900A,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_nY,	0,		0},
	{0xD208,0x8008,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Yq,	0x2C07,	0},
	{0xFE0F,0x8000,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Z,	0,		0},
	{0xFE0F,0x9001,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Zp,	0,		0},
	{0xFE0F,0x9002,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_nZ,	0,		0},
	{0xD208,0x8000,"LD",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Zq,	0x2C07,	0},
	{0xF000,0xE000,"LDI",		0,	2,	2, 	0,	P_Rd,	0x00F0,	P_K,	0x0F0F,	0},
	{0xFE0F,0x9000,"LDS",		0,	4,	2,	0,	P_Rd,	0x01F0,	P_kk,	0,		0},
	{0xF800,0xA000,"LDS",		0,	2,	2,	0,	P_Rd,	0x00F0, P_kk,	0x070F,	0}, //4to nado delatj P_kk dlja adresa
	{0xFFFF,0x95C8,"LPM",		0,	2,  0,	0,	P_NONE, 0,		P_NONE,	0,		0},
	{0xFE0F,0x9004,"LPM",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_Z, 	0,		0},
	{0xFE0F,0x9005,"LPM",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_Zp,	0,		0},//{0xFC00,0x0C00,"LSL",		0,	2,	1,	0,	P_Rd,	0x03FF,	P_NONE,	0,		0},Rd=Rd+Rd
	{0xFE0F,0x9406,"LSR",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE, 0,		0},
	{0xFC00,0x2C00,"MOV",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_NONE,	0x20F,	0},
	{0xFF00,0x0100,"MOVW",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_Rr,	0x000F,	0},
	{0xFC00,0x9C00,"MUL",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xFF00,0x0200,"MULS",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_Rr,	0x000F,	0},
	{0xFF88,0x0300,"MULSU",		0,	2,	2,	0,	P_Rd,	0x0070,	P_Rr,	0x0007,	0},
	{0xFE0F,0x9401,"NEG",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFFFF,0x0000,"NOP",		0,	2,	2,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFC00,0x2800,"OR",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xF000,0x6000,"ORI",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_K,	0x0F0F,	0},
	{0xF800,0xB800,"OUT",		0,	2,	2,	0,	P_A,	0x060F,	P_Rr,	0x01F0,	0},
	{0xFE0F,0x900F,"POP",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFE0F,0x920F,"PUSH",		0,	2,	1,	0,	P_Rr,	0x01F0,	P_NONE,	0,		0},
	{0xF000,0xD000,"RCALL",		0,	2,	1,	0,	P_k,	0x0FFF,	P_NONE,	0,		0},
	{0xFFFF,0x9508,"RET",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9518,"RETI",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xF000,0xC000,"RJMP",		0,	2,	1,	0,	P_k,	0x0FFF,	P_NONE, 0,		0},//ROL is ADC
	{0xFE0F,0x9407,"ROR",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE, 0,		0},
	{0xFC00,0x0800,"SBC",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x020F,	0},
	{0xF000,0x4000,"SBCI",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_K,	0x0F0F,	0},
	{0xFF00,0x9A00,"SBI",		0,	2,	2,	0,	P_A,	0x00F8,	P_b,	0x0007,	0},
	{0xFF00,0x9900,"SBIC",		0,	2,	2,	0,	P_A,	0x00F8,	P_b,	0x0007,	0},
	{0xFF00,0x9B00,"SBIS",		0,	2,	2,	0,	P_A,	0x00F8,	P_b,	0x0007,	0},
	{0xFF00,0x9700,"SBIW",		0,	2,	2,	0,	P_Rd,	0x0030,	P_K,	0x00CF, 0},
	{0xF000,0x6000,"SBR",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_K,	0x0F0F, 0},
	{0xFE08,0xFC00,"SBRC",		0,	2,	2,	0,	P_Rr,	0x01F0,	P_b,	0x0007,	0},
	{0xFE08,0xFE00,"SBRS",		0,	2,	2,	0,	P_Rr,	0x01F0,	P_b,	0x0007,	0},
	{0xFFFF,0x9408,"SEC",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9458,"SEH",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9478,"SEI",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9428,"SEN",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFF0F,0xEF0F,"SER",		0,	2,	1,	0,	P_Rd,	0x00F0,	P_NONE,	0,		0},
	{0xFFFF,0x9448,"SES",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9468,"SET",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9438,"SEV",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9418,"SEZ",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x9588,"SLEEP",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x95E8,"SPM",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	{0xFFFF,0x95F8,"SPM",		0,	2,	1,	0,	P_Zp,	0,		P_NONE,	0,		0},
	{0xFE0F,0x920C,"ST",		0,	2,	2,	0,	P_X,	0,		P_Rr,	0x01F0,	1},
	{0xFE0F,0x920D,"ST",		0,	2,	2,	0,	P_Xp,	0,		P_Rr,	0x01F0,	1},
	{0xFE0F,0x920E,"ST",		0,	2,	2,	0,	P_nX,	0,		P_Rr,	0x01F0,	1},
	{0xFE0F,0x8208,"ST",		0,	2,	2,	0,	P_Y,	0,		P_Rr,	0x01F0,	0},
	{0xFE0F,0x9209,"ST",		0,	2,	2,	0,	P_Yp,	0,		P_Rr,	0x01F0,	0},
	{0xFE0F,0x920A,"ST",		0,	2,	2,	0,	P_nY,	0,		P_Rr,	0x01F0,	0},
	{0xD208,0x8208,"ST",		0,	2,	2,	0,	P_Yq,	0x2C07,	P_Rr,	0x01F0,	0},	
	{0xFE0F,0x8200,"ST",		0,	2,	2,	0,	P_Z,	0,		P_Rr,	0x01F0,	0},
	{0xFE0F,0x8201,"ST",		0,	2,	2,	0,	P_Zp,	0,		P_Rr,	0x01F0,	0},
	{0xFE0F,0x8202,"ST",		0,	2,	2,	0,	P_nZ,	0,		P_Rr,	0x01F0,	0},
	{0xD207,0x8200,"ST",		0,	2,	2,	0,	P_Zq,	0x2C07,	P_Rr,	0x01F0,	0},
	{0xFE0F,0x9200,"STS",		0,	4,	2,	0,	P_kk,	0,		P_Rr,	0x01F0,	0},
	{0xF800,0xA800,"STS",		0,	2,	2,	0,	P_k,	0x070F,	P_Rr,	0x00F0,	0},
	{0xFC00,0x1800,"SUB",		0,	2,	2,	0,	P_Rd,	0x01F0,	P_Rr,	0x02F0,	0},
	{0xF000,0x5000,"SUBI",		0,	2,	2,	0,	P_Rd,	0x00F0,	P_K,	0x0F0F,	0},
	{0xFE0F,0x9402,"SWAP",		0,	2,	1,	0,	P_Rd,	0x01F0,	P_NONE,	0,		0},
	{0xFFFF,0x95A8,"WDR",		0,	2,	0,	0,	P_NONE,	0,		P_NONE,	0,		0},
	//{0x0,	0x0,	0x0,		0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},
};

uint16_t get_opcode_num( uint16_t data )
{
	int i;
	//for (i=0; i<OPCODE_NUMBER; i++)
	for (i=0; opcode_table[i].size != 0x0; i++)
	{
		if ((data&opcode_table[i].mask) == opcode_table[i].opcode)
			return i;
	}
	return 0xFFFF;
}

void opcode_print( uint16_t instr , uint16_t data1, uint16_t data2 )
{
	if ((instr >= 0) and (instr < OPCODE_NUMBER))
	{
		printf( "%04x " , data1 );
		switch (opcode_table[instr].param_num)
		{
			case 0:
				if (opcode_table[instr].size == 0x0002)
				{
					printf("     %-5s\n",opcode_table[instr].name);
				}
				if (opcode_table[instr].size == 0x0004)
				{
					printf("0,2\n");
				}
				break;
			case 1:
				if (opcode_table[instr].size == 0x0004)
				{
					uint32_t addr=0x0;
					addr = ((((data1&opcode_table[instr].param_mask1)>>4) + (data1&0x1))<<16)+(data2);
					//addr = data2;
					
					printf("%04d %-5s 0x%-8x\n",data2,opcode_table[instr].name,addr<<1);
				}
				if (opcode_table[instr].size == 0x0002)
				{
					uint16_t param=0x0;
					param = get_mask( data1 , opcode_table[instr].param_mask1 );
					
					if ((strncmp(opcode_table[instr].name,"RJMP",4)==0)
						or (strncmp(opcode_table[instr].name,"RCALL",5) == 0))
					{
						if ((param&0x0800) == 0x800)
						{
							printf("     %-5s -%-04d\n", opcode_table[instr].name, (0xFFF-param+1)*2);
						}else
						{
							printf("     %-5s +%-04d \n", opcode_table[instr].name, (param)*2);
						}
						break;
					}
					
					printf("     %-5s 0x%-4x\n", opcode_table[instr].name, param);
					
				}
				break;
			case 2:
				if (opcode_table[instr].size == 0x0002)
				{
					uint16_t p1=0,p2=0;
					p1 = get_mask( data1, opcode_table[instr].param_mask1);
					p2 = get_mask( data1, opcode_table[instr].param_mask2);
					if ( strncmp(opcode_table[instr].name,"BRBC",4) == 0)
					{
						if (((p2 & 0x0040) == 0x040))
						{
							printf("     %-5s 0x%-04x , -%-02d \n", opcode_table[instr].name , p1 , (128-p2)*2 );
						} else
						{
							printf("     %-5s 0x%-04x , +%-02d \n", opcode_table[instr].name , p1 , 1 );
						}
						break;
					}
					printf("     %-5s 0x%-4x , 0x%-4x \n", opcode_table[instr].name , p1 , p2 );
				}
				if (opcode_table[instr].size == 0x0004)
				{
					printf("2,2\n");
				}
				break;
			default:
				printf("Unknown param number\n");
				break;
		}
	}
}

uint16_t get_mask( uint16_t data, uint16_t mask)
{
	uint16_t m=0x0;
	switch ( mask )
	{
		case 0x0000:
			m = 0x0000;
			break;
		case 0x01F0:
			m = (data&mask) >> 4;
			break;
		case 0x0070:
			m = (data&mask) >> 4;
			break;
		case 0x03FF:
			m = (data&mask);
			break;
		case 0x00F0:
			m = (data&mask)>>4;
			break;
		case 0x0FFF:
			m = (data&mask);
			break;
		case 0x060F:
			m = ((data&0x0600)>>5) + (data&0x000F);
			break;
		case 0x020F:
			m = ((data&0x0200)>>5) + (data&0x000F);
			break;
		case 0x0030:
			m = (data&0x0030)>>4;
			break;
		case 0x00CF:
			m = ((data&0x00C0)>>2) + (data&0x000F);
			break;
		case 0x0F0F:
			m = ((data&0x0F00)>>4)+(data&0x000F);
			break;
		case 0x0007:
			m = (data&0x0007);
			break;
		case 0x00F8:
			m = (data&0x00F8)>>3;
			break;
		case 0x2C07:
			m = ((data&0x2000)>>13) + ((data&0x0C00)>>10) + (data&0x0007);
			break;
		case 0x01F1:
			m = ((data&0x01F0)>>4)+(data&0x0001);
			break;
		case 0x070F:
			m = ((data&0x0700)>>8) + (data&0x000F);
			break;
		case 0x03F8:
			m = (data&0x03f8)>>3;
			break;
		default:
			m=0xFFFF;
	}
	return m;
}
